///****************************   LICENSE   *****************************
/// Copyright (C) 2019, HUSEYIN TOKPUNAR http://huseyintokpinar.com/
/// Download & Update Latest Version: https://github.com/hhtokpinar/sqfEntity
/// Licensed under the Apache License, Version 2.0 (the "License");
/// you may not use this file except in compliance with the License.
/// You may obtain a copy of the License at

///    http://www.apache.org/licenses/LICENSE-2.0

/// Unless required by applicable law or agreed to in writing, software
/// distributed under the License is distributed on an "AS IS" BASIS,
/// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
/// See the License for the specific language governing permissions and
/// limitations under the License.
///**********************************************************************/

part of 'sqfentity_gen.dart';

class Controllers {
  static List<String> controllersub = List<String>();
  static List<String> controllers = List<String>();
}

class SqfEntityFormConverter {
  SqfEntityFormConverter(this.table);
  final SqfEntityTableBase table;
  String toFormWidgetsCode() {
    final String tablename = table.tableName.toLowerCase();
    final String modelName = table.modelName ?? toCamelCase(table.tableName);

    String getformListTitleField(SqfEntityTableBase table) {
      String fieldName = '';
      if (table.formListTitleField != null) {
        fieldName = table.formListTitleField;
      } else {
        for (final field in table.fields) {
          if (field.dbType == DbType.text) {
            fieldName = field.fieldName;
            break;
          }
        }
        if (fieldName.isEmpty) {
          fieldName = table.primaryKeyName;
        }
      }
      return fieldName;
    }

    String subControllers() {
      final retVal = StringBuffer();
      for (final field in table.fields) {
        if (field is SqfEntityFieldRelationshipBase) {
          final objName = field.relationshipName == null
              ? table.modelName
              : field.relationshipName;
          if (!Controllers.controllersub.contains(objName)) {
            retVal.writeln(
                '''class ${field.relationshipName}ControllerSub extends ${modelName}Controller {
                 static String relationshipFieldName='${field.fieldName}';
                 static String primaryKeyName= '${table.primaryKeyName}';
                 static bool useSoftDeleting = ${table.useSoftDeleting.toString()};
                 static String formListTitleField = '${getformListTitleField(table)}';
                 static String formListSubTitleField${table.formListSubTitleField == null ? '' : ' = \''}${table.formListSubTitleField == null ? '' : table.formListSubTitleField}${table.formListSubTitleField == null ? '' : '\''};
                 }''');
            Controllers.controllersub.add(objName);
          }
        }
      }
      return retVal.toString();
    }

    final String subItemsCode = table.hasSubItems
        ? '''SQFViewList subList(int id) {
    return SQFViewList(${modelName}ControllerSub(),
      primaryKeyName: ${modelName}ControllerSub.primaryKeyName,
      useSoftDeleting: ${modelName}ControllerSub.useSoftDeleting,
      formListTitleField: ${modelName}ControllerSub.formListTitleField,
      formListSubTitleField: ${modelName}ControllerSub.formListSubTitleField,
      filterExpression: '\${${modelName}ControllerSub.relationshipFieldName}=\$id');
  }'''
        : '';

    return '''
    // CONTROLLER hasSubItems:${table.hasSubItems}
    ${subControllers().toString()}
    class ${modelName}Controller extends $modelName{

     static SQFViewList getController = SQFViewList(
      ${table.modelName}Controller(),
      primaryKeyName: '${table.primaryKeyName}',
      useSoftDeleting: ${table.useSoftDeleting.toString()},
      formListTitleField: '${getformListTitleField(table)}',
      formListSubTitleField: ${table.formListSubTitleField == null ? '' : '\''}${table.formListSubTitleField == null ? 'null' : table.formListSubTitleField}${table.formListSubTitleField == null ? '' : '\''},
      hasSubItems: ${table.hasSubItems},
    );
    $subItemsCode
    Future<Widget> gotoEdit(dynamic obj) async{
    return ${modelName}Add(obj==null ? $modelName() : await $modelName().getById(obj['${table.primaryKeyName}']${table.primaryKeyType == PrimaryKeyType.text ? '.toString()' : ' as int'}) ?? $modelName());
  }
}
// END CONTROLLER

class ${modelName}Add extends StatefulWidget {
  ${modelName}Add(this._$tablename);
  final dynamic _$tablename;
  @override
  State<StatefulWidget> createState() => ${modelName}AddState(_$tablename as $modelName);
}

class ${modelName}AddState extends State {
  ${modelName}AddState(this.$tablename);
  $modelName $tablename;
  final _formKey = GlobalKey<FormState>();
  ${toFormDeclarationCodeTable(table).toString()}

  @override
  void initState() {
    ${toFormInitStateCodeTable(table).toString()}
    super.initState();
  }

  @override
  Widget build(BuildContext context) {
    ${toFormBuildDropDownCodeTable(table).toString()}

    return Scaffold(
      appBar: AppBar(
        title: ($tablename.${table.primaryKeyName} == null)
            ? Text('Add a new $tablename')
            : Text('Edit $tablename'),
      ),
      body: Container(
        alignment: Alignment.topCenter,
        width: double.infinity,
        height: double.infinity,
        child: SingleChildScrollView(
          child: Padding(
              padding: EdgeInsets.only(top: 10.0, left: 10.0, right: 10.0),
              child: Form(
                key: _formKey,
                child: Column(
                  children: <Widget>[
                    ${toFormBuildRowWidgets(table).toString()}
                    FlatButton(
                      child: saveButton(),
                      onPressed: () {
                        if (_formKey.currentState.validate()) {
                          // If the form is valid, display a Snackbar.
                          save();
                          Scaffold.of(context).showSnackBar(SnackBar(
                              duration: Duration(seconds: 2),
                              content: Text('Processing Data')));
                        }
                      },
                    )
                  ],
                ),
              )),
        ),
      ),
    );
  }

  ${toFormBuildRowCodeTable(table).toString()}

  
  Container saveButton() {
    return Container(
      padding: const EdgeInsets.all(7.0),
      decoration: BoxDecoration(
          color: Color.fromRGBO(95, 66, 119, 1.0),
          border: Border.all(color: Colors.black),
          borderRadius: BorderRadius.circular(5.0)),
      child: Text(
        'Save',
        style: TextStyle(color: Colors.white, fontSize: 20),
      ),
    );
  }

  void save() async {
    ${toFormSaveCodeVariables(table).toString()}
    $tablename
      ${toFormSaveCode(table).toString()};
    final result = await $tablename.save();
    if (result ${table.primaryKeyType == PrimaryKeyType.text ? '.success' : ' != 0'}) {
      Navigator.pop(context, true);
    } else
    {
      UITools(context).alertDialog($tablename.saveResult.toString(),
                  title: 'save ${table.modelName} Failed!', callBack: () {
                Navigator.pop(context, true);
              });
    }
  }
}
''';
  }

  String toFormBuildRowWidgets(SqfEntityTableBase table) {
    final retVal = StringBuffer();
    for (final field in table.fields.where((f) => f.sequencedBy == null)) {
      final ccName = toCamelCase(field.fieldName);
      if (field is SqfEntityFieldRelationshipBase) {
        retVal.writeln('buildRow$ccName(onChangeDropdownItemFor$ccName),');
      } else {
        retVal.writeln('buildRow$ccName(),');
      }
    }
    return retVal.toString();
  }

  String toFormSaveCodeVariables(SqfEntityTableBase table) {
    final retVal = StringBuffer();
    for (final field in table.fields
        .where((f) => f.dbType == DbType.date || f.dbType == DbType.datetime)) {
      final ccName = toCamelCase(field.fieldName);
      final vName = tocamelCase(field.fieldName);
      retVal.writeln('var _$vName = DateTime.tryParse(txt$ccName.text);');
      if (field.dbType == DbType.datetime) {
        retVal.writeln(
            '''final _${vName}Time = DateTime.tryParse(txtTimeFor$ccName.text);
    if (_$vName != null && _${vName}Time != null) {
      _$vName = _$vName.add(
          Duration(hours: _${vName}Time.hour, minutes: _${vName}Time.minute, seconds: _${vName}Time.second));
    }''');
      }
    }
    return retVal.toString();
  }

  String toFormSaveCode(SqfEntityTableBase table) {
    final retVal = StringBuffer();
    for (final field in table.fields.where((f) => f.dbType != DbType.bool)) {
      final ccName = toCamelCase(field.fieldName);
      if (field is SqfEntityFieldRelationshipBase) {
        retVal.writeln('..${field.fieldName} = _selected$ccName');
      } else {
        switch (field.dbType) {
          case DbType.date:
          case DbType.datetime:
            retVal.writeln(
                '..${field.fieldName} = _${tocamelCase(field.fieldName)}');
            break;
          case DbType.integer:
          case DbType.numeric:
            retVal.writeln(
                '..${field.fieldName} = int.tryParse(txt$ccName.text)');
            break;
          case DbType.real:
            retVal.writeln(
                '..${field.fieldName} = double.tryParse(txt$ccName.text)');
            break;
          default:
            retVal.writeln('..${field.fieldName} = txt$ccName.text');
        }
      }
    }
    return retVal.toString();
  }

  String toFormBuildRowCodeTable(SqfEntityTableBase table) {
    final retVal = StringBuffer();
    for (final field in table.fields) {
      retVal.writeln(toFormBuildRowCodeField(field));
    }
    return retVal.toString();
  }

  String toFormDeclarationCodeTable(SqfEntityTableBase table) {
    final retVal = StringBuffer();
    for (final field in table.fields) {
      retVal.writeln(toFormDeclarationCodeField(field));
    }
    return retVal.toString();
  }

  String toFormInitStateCodeTable(SqfEntityTableBase table) {
    final retVal = StringBuffer();
    for (final field in table.fields) {
      if (field is SqfEntityFieldRelationshipBase) {
      } else {
        retVal.writeln(toFormInitStateCodeField(field));
      }
    }
    return retVal.toString();
  }

  String toFormDeclarationCodeField(SqfEntityFieldType field) {
    final String ccName = toCamelCase(field.fieldName);
    if (field is SqfEntityFieldRelationshipBase) {
      return '''List<DropdownMenuItem<int>> _dropdownMenuItemsFor$ccName =
      List<DropdownMenuItem<int>>();
  int _selected$ccName;
        ''';
    } else {
      switch (field.dbType) {
        case DbType.bool:
          return '';
        case DbType.date:
          return 'final txt$ccName = TextEditingController();';
        case DbType.datetime:
          return 'final txt$ccName = TextEditingController(); final txtTimeFor$ccName = TextEditingController();';
        default:
          return 'final txt$ccName = TextEditingController();';
      }
    }
  }

  String toFormInitStateCodeField(SqfEntityFieldType field) {
    final String objName = field.table.tableName.toLowerCase();
    final String fName = field.fieldName;
    final String ccName = toCamelCase(fName);
    switch (field.dbType) {
      case DbType.real:
      case DbType.integer:
      case DbType.numeric:
        return 'txt$ccName.text = $objName.$fName == null? \'\': $objName.$fName.toString();';
      case DbType.bool:
        return '';
      case DbType.date:
        return 'txt$ccName.text = $objName.$fName == null? \'\': UITools.convertDate($objName.$fName);';
      case DbType.datetime:
        return '''txt$ccName.text = $objName.$fName == null? \'\': UITools.convertDate($objName.$fName);
        txtTimeFor$ccName.text = $objName.$fName == null? \'\': UITools.convertTime($objName.$fName);
        ''';
      default:
        return 'txt$ccName.text = $objName.$fName == null? \'\': $objName.$fName;';
    }
  }

  String toFormBuildDropDownCodeTable(SqfEntityTableBase table) {
    final retVal = StringBuffer();
    for (final field in table.fields) {
      if (field is SqfEntityFieldRelationshipBase) {
        field.relationshipName = field.relationshipName ?? table.modelName;
        retVal.writeln(toFormBuildDropDownCodeField(field));
      }
    }
    return retVal.toString();
  }

  String toFormBuildDropDownCodeField(SqfEntityFieldRelationshipBase field) {
    final ccName = toCamelCase(field.fieldName);
    return '''void buildDropDownMenuFor$ccName() async {
      final dropdownMenuItems =
          await ${field.relationshipName}().select().toDropDownMenuInt('${field.formDropDownTextField ?? getformListTitleField(field.table)}');
      setState(() {
        _dropdownMenuItemsFor$ccName = dropdownMenuItems;
        _selected$ccName = ${table.tableName.toLowerCase()}.${field.fieldName};
      });
    }
    if (_dropdownMenuItemsFor$ccName == null ||
        _dropdownMenuItemsFor$ccName.isEmpty) {
      buildDropDownMenuFor$ccName();
    }
    void onChangeDropdownItemFor$ccName(int selected$ccName) {
      setState(() {
        _selected$ccName = selected$ccName;
      });
    }''';
  }

  String toFormBuildRowCodeField(SqfEntityFieldType field) {
    final ccName = toCamelCase(field.fieldName);
    final tablename = field.table.tableName.toLowerCase();
    final retVal = StringBuffer()..write('''Widget buildRow$ccName() {''');

    if (field is SqfEntityFieldRelationshipBase) {
      return '''Widget buildRow$ccName(void onChangeDropdownItemFor$ccName(int selected$ccName)) {
    return Row(
      children: <Widget>[
        Expanded(
          flex: 1,
          child: Text('${field.relationshipName}'),
        ),
        Expanded(
            flex: 1,
            child: Container(
              child: DropdownButtonFormField(
                value: _selected$ccName,
                items: _dropdownMenuItemsFor$ccName,
                onChanged: onChangeDropdownItemFor$ccName,
                validator: (value) {
                  if ((_selected$ccName != null && _selected$ccName > 0) ${field.formIsRequired != null && field.formIsRequired ? '' : '|| true'}) {
                    return null;
                  } else if (value == null || value == 0) {
                    return 'Please enter ${field.relationshipName}';
                  }
                  return null;
                },
              ),
            )),
      ],
    );
  }''';
    }

    switch (field.dbType) {
      case DbType.integer:
      case DbType.numeric:
      case DbType.real:
        final rangeValidator = StringBuffer();
        if (field.minValue != null && field.maxValue != null) {
          rangeValidator.write(
              '''else if(${field.dbType == DbType.real ? 'double' : 'int'}.tryParse(value) >${field.maxValue} || ${field.dbType == DbType.real ? 'double' : 'int'}.tryParse(value) < ${field.minValue}){ 
              return 'Please Enter Between ${field.minValue} and ${field.maxValue} (required)'; }''');
        } else if (field.minValue != null) {
          rangeValidator.write('''else if(${field.dbType == DbType.real ? 'double' : 'int'}.tryParse(value) < ${field.minValue}){ 
              return 'Please Enter Minimum ${field.minValue} (required)'; }''');
        } else if (field.maxValue != null) {
          rangeValidator.write('''else if(${field.dbType == DbType.real ? 'double' : 'int'}.tryParse(value) > ${field.maxValue}){ 
              return 'Please Enter Maximum ${field.maxValue} (required)'; }''');
        }

        if ((field.formIsRequired != null && field.formIsRequired)) {


          retVal.writeln('''return TextFormField(
      validator: (value) {
         if (${field.dbType == DbType.real ? 'double' : 'int'}.tryParse(value) == null) 
            { 
              return 'Please Enter valid number (required)'; 
            }
        ${rangeValidator.toString()}
        return null;
      },''');
        } else {
          retVal.writeln('''return TextFormField(
      validator: (value) { if (value.isNotEmpty && ${field.dbType == DbType.real ? 'double' : 'int'}.tryParse(value) == null) 
      { return 'Please Enter valid number'; }
      ${rangeValidator.toString().replaceAll('(required)', '')}
        return null;
      },''');
        }
        break;
      case DbType.bool:
        retVal.writeln('''return Row(
      children: <Widget>[
        Text('$ccName?'),
        Checkbox(
          value: $tablename.${field.fieldName} ?? false,
          onChanged: (bool value) {
            setState(() {
              $tablename.${field.fieldName} = value;
            });
          },
        ),
        
      ],
    );
  }
''');
        return retVal.toString();
      case DbType.date:
        retVal.writeln('return TextFormField(');
        retVal.writeln(
            '''onTap: () => DatePicker.showDatePicker(context, showTitleActions: true, theme: UITools.mainDatePickerTheme
          ${_getNullableValue(field.minValue, 'minTime')}
          ${_getNullableValue(field.maxValue, 'maxTime')}
          , onConfirm: (sqfSelectedDate) {
        txt$ccName.text = UITools.convertDate(sqfSelectedDate);
        setState(() {
          $tablename.${field.fieldName} = sqfSelectedDate;
        });
      }, currentTime: DateTime.tryParse(txt$ccName.text) ?? $tablename.${field.fieldName} ?? DateTime.now(), locale: UITools.mainDatePickerLocaleType),
      ''');
        break;
      case DbType.datetime:
        retVal.writeln(
            '''return Row(children: <Widget>[Expanded(flex: 1, child: TextFormField(onTap: () => DatePicker.showDatePicker(context, showTitleActions: true, theme: UITools.mainDatePickerTheme
          ${_getNullableValue(field.minValue, 'minTime')}
          ${_getNullableValue(field.maxValue, 'maxTime')}
          , onConfirm: (sqfSelectedDate) {
        txt$ccName.text = UITools.convertDate(sqfSelectedDate);
        txtTimeFor$ccName.text = UITools.convertTime(sqfSelectedDate);
            setState(() {
                final d = DateTime.tryParse(txt$ccName.text) ?? $tablename.${field.fieldName} ?? DateTime.now();
                $tablename.${field.fieldName} = DateTime(sqfSelectedDate.year, sqfSelectedDate.month, sqfSelectedDate.day).add(Duration(hours: d.hour,minutes: d.minute,seconds: d.second));
            });
          },
              currentTime: DateTime.tryParse(txt$ccName.text) ?? $tablename.${field.fieldName} ?? DateTime.now(),
              locale: UITools.mainDatePickerLocaleType),
          ${field.formIsRequired != null && field.formIsRequired ? 'validator: (value) { if (value.isEmpty) { return \'Please enter $ccName\'; } return null;},' : ''}
          controller: txt$ccName,
          decoration: InputDecoration(labelText: '$ccName'),
        ),
      ),
      Expanded(flex: 1,child: TextFormField(onTap: () => DatePicker.showTimePicker(context,showTitleActions: true,theme: UITools.mainDatePickerTheme, 
      onConfirm: (sqfSelectedDate) {
              txtTimeFor$ccName.text = UITools.convertTime(sqfSelectedDate);
              setState(() {
                final d = DateTime.tryParse(txt$ccName.text) ?? $tablename.${field.fieldName} ?? DateTime.now();
                $tablename.${field.fieldName} = DateTime(d.year, d.month, d.day).add(Duration(hours: sqfSelectedDate.hour,minutes: sqfSelectedDate.minute,seconds: sqfSelectedDate.second));
                txt$ccName.text = UITools.convertDate($tablename.${field.fieldName});
              });
            },
                currentTime: DateTime.tryParse('\${UITools.convertDate(DateTime.now())} \${txtTimeFor$ccName.text}') ?? $tablename.${field.fieldName} ?? DateTime.now(),
                locale: UITools.mainDatePickerLocaleType),
            controller: txtTimeFor$ccName,
            decoration: InputDecoration(labelText: ''),
          ))
    ]);
}''');
        return retVal.toString();
        break;
      default:
        if ((field.formIsRequired != null && field.formIsRequired)) {
          retVal.writeln('''return TextFormField(
     validator: (value) { if (value.isEmpty)
     { return 'Please enter $ccName'; }
        return null;
     },''');
        } else {
          retVal.writeln('return TextFormField(');
        }
    }

    retVal.writeln('''      controller: txt$ccName,
      decoration: InputDecoration(labelText: '$ccName'),
    );
  }''');
    return retVal.toString();
  }
}

String getformListTitleField(SqfEntityTableBase table) {
  String fieldName = '';
  if (table.formListTitleField != null) {
    fieldName = table.formListTitleField;
  } else {
    for (final field in table.fields) {
      if (field.dbType == DbType.text) {
        fieldName = field.fieldName;
        break;
      }
    }
    if (fieldName.isEmpty) {
      fieldName = table.primaryKeyName;
    }
  }
  return fieldName;
}
