// Copyright (c) 2018, the Dart project authors. Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

import 'package:analyzer/dart/element/element.dart';
import 'package:build/build.dart';
import 'package:source_gen/source_gen.dart';
import 'package:sqfentity_gen/src/sqfentity_helper.dart';
import '../sqfentity_gen.dart';

//import 'package:sqfentity_base/sqfentity_base.dart';

class Imports {
  static Map<String, bool> importedModels = Map<String, bool>();
  static List<String> controllers = List<String>();
}

class SqfEntityFormGenerator
    extends GeneratorForAnnotation<SqfEntityBuilderForm> {
  @override
  String generateForAnnotatedElement(
      Element element, ConstantReader annotation, BuildStep buildStep) {
    final tablemodel = annotation.read('table').objectValue;
    final formListTitleField =
        getStringValueAnnotation(annotation, 'formListTitleField');
    final formListSubTitleField =
        getStringValueAnnotation(annotation, 'formListSubTitleField');
    final bool hasSubItems =
        getBoolValueAnnotation(annotation, 'hasSubItems') ?? false;

    // When testing, you can uncomment the test line to make sure everything's working properly
    //return '// MODEL -> ${tablemodel.toString()}';

    var instanceName =
        'SqfEntityTable'; //        element.toString().replaceAll('SqfEntityTable', '').trim();
    instanceName = toCamelCase(instanceName);

    final builder = SqfEntityFormCreator(tablemodel, instanceName,
        formListTitleField, formListSubTitleField, hasSubItems);

    print(
        '-------------------------------------------------------FormBuilder: $instanceName');
    final table = builder.toTable;
    final String path = element.source
        .toString()
        .substring(element.source.toString().lastIndexOf('/') + 1);
    if (Imports.controllers.contains(table.tableName)) {
      final String errMessage = 'SQFENTITY_FORMGENERATOR ERROR: ${table.tableName} MODEL ALREADY EXIST. Please check $path file for dublicate annotation error';
      print(errMessage);
      throw Exception(errMessage);
      //return '';
    }
    Imports.controllers.add(table.tableName);
    print('${table.modelName} Model recognized succesfuly');
    final modelStr = StringBuffer()

          //..writeln('/*') // write output as commented to see what is wrong
          ..writeln(SqfEntityFormConverter(table).toFormWidgetsCode())
        //..writeln('*/') //  write output as commented to see what is wrong
        ;

    if (Imports.importedModels[path] != null) {
      return modelStr.toString();
    } else {
      Imports.importedModels[path] = true;
      return 'part of \'$path\';  ${modelStr.toString()}';
    }
  }
}
